{% extends "timeline-page.html" %}

{% block stylesheets %}
    <link rel="stylesheet" href="/static/css/consent.css"/>
{% endblock %}

{% block main_body %}
<style>
  .center {
    display: block;
    margin-left: auto;
    margin-right: auto;
    width: 50%;
  }
</style>

<img src="/static/images/princeton-consent.png" alt="Princeton University adult consent form" style="width:450px" class="center">
<div class="main-div" style="padding-bottom: 100px; font-size: 18px;">
  <div class="text-justify">
    <p>
        TITLE OF RESEARCH: <i>Studying Perceptual Biases</i><br>
        PRINCIPAL INVESTIGATOR: <i>Thomas Griffiths</i><br>
        PRINCIPAL INVESTIGATOR’S DEPARTMENT: <i>Psychology</i>
    </p>

    <p>
        You are being invited to take part in a research study. Before you decide to participate in this study, it is
        important that you understand why the research is being done and what it will involve. Please take the
        time to read the following information carefully. Please ask the researcher if there is anything that is not
        clear or if you need more information.
    </p>

    <h2>Purpose of the research:</h2>
    <p>
        The purpose of this project is to study how people from different places in the world experience sounds
        (e.g., speech, music) and images (e.g., faces, color).
    </p>

    <h2>Study Procedures:</h2>
    <p>
        If you volunteer to participate in this study, you will be asked to perform simple tasks using your keyboard
        or mouse while listening to sounds, or while observing images or videos. For example, you may be asked to judge
        how pleasant a particular sound or an image is, or to categorize sounds or images (e.g., if a sound or an image
        is that of a car or a cow), or to judge other basic features of sounds or images (e.g., how frequently do you
        encounter this type of sound or image? how aesthetically pleasing you consider a particular sound or image?).
        The specific sounds, images, and videos used will vary, but may include: recordings of ambient sounds, noise-bursts,
        speech, music, simple tones, naturalistic images, animated images, comics, faces, objects, geometric shapes,
        fractal images, or videos of natural environments.
    </p>
    <p>
        In some experiments, we may use mild deception, i.e., we may not explain the full truth regarding an experimental
        task (for example we will
        not tell you that an image was generated by an artificial computer system and not a human). Mild deception will
        only be used in some experiments where we need to prevent biasing the participants before taking the experiment.
    </p>
    <p>
        Finally, <span style="font-weight: bold;">in some experiments you may be asked to make a voice or video recording</span>, for example, you may
        be asked to record your finger-tapping in response to a sound or video, or to record yourself speaking or singing.
        If the experiment will contain these elements we will ask for an additional consent.
    </p>
    <p>
        Your participation in this research is voluntary. You are free to refuse to take part, and you may stop taking
        part at any time. You are free to discontinue participation in this study at any time without penalty. The
        investigator may withdraw you from this research if circumstances arise which warrant doing so.
    </p>
    <p>
        The duration of the experiment will range from 10-90 minutes.
    </p>

    <h2>Benefits and Risks:</h2>
    <p>
        There are no direct benefits to you as a participant; however, by furthering our understanding of human cognition,
        this research will benefit society by helping understand explore the evolutionary causes and consequences
        of human perception.<br>Risks associated with participation in this study are minimal. You may feel slight discomfort
        answering some questions, but you may refrain from answering any questions that make you uncomfortable and may
        withdraw your participation at any time, even after completing the experiment, without penalty.
    </p>

    <h2>Confidentiality:</h2>
    <p>
        We will not be asking for any personally identifying information, and we will handle responses as
        confidentially as possible. Worker IDs will never be tied to your responses on this survey. However,
        we cannot guarantee the confidentiality of information transmitted over the Internet. To minimize
        this risk, data containing anything that might be personally identifiable (e.g. Worker IDs or IP
        addresses) will be encrypted on transfer and storage and will only be accessible to qualified lab
        personnel. We will be keeping data collected as part of this experiment indefinitely. This anonymized
        data (containing neither Worker IDs nor IP addresses) may be shared with the scientific community
        or with other participants to be used as stimuli in future studies.
    </p>
    <p>
        No personal identifiers (e.g. your name or contact data) will be collected or used at any stage
        of the experiment. Thus, the experimental data is anonymous. We may collect some demographic data
        (e.g. age, gender, country of residence) as well as information about your learning and exposure
        history (e.g. musical experience playing an instrument, years of formal education, languages spoken).
    </p>
    <p>
        For the purpose of helping scientific progress and replication, sometimes we may share the data
        with research collaborators. In these cases, we will only share fully anonymized data, such as
        demographic, aggregated, or extracted data without any personal identifiers.
    </p>

    <h2>Compensation:</h2>
    <p>
        For your participation, you will be paid in accordance with the terms of the survey platform.
        If you are an online participant and for any reason you do not complete the study (e.g. technical
        difficulties, or a desire to stop), we will only be able to pay you if you email the researcher,
        Raja Marjieh, at <a href="mailto:cocosci-lab@princeton.edu">cocosci-lab@princeton.edu</a>. If you
        have any questions about the study, feel free to contact Raja Marjieh or the Principal Investigator,
        Thomas Griffiths, at <a href="mailto:tomg@princeton.edu">tomg@princeton.edu</a>.
    </p>

    <h2>Who to contact with questions:</h2>
    <p>
        <ol>
            <li>PRINCIPAL INVESTIGATOR:<br><br>
                Dr. Thomas Griffiths<br>
                tomg@princeton.edu <br><br>
            </li>
            <li>If you have questions regarding your rights as a research subject, or if problems arise which you do
                not feel you can discuss with the Investigator, please contact the Institutional Review Board at:<br><br>
                Assistant Director, Research Integrity and Assurance<br>
                Phone: (609) 258-8543,<br>
                Email: irb@princeton.edu<br><br>
            </li>
            <li>
                I understand the information that was presented and that:
                <ol type="A">
                    <li>
                        My participation is voluntary, and I may withdraw my consent and discontinue participation in the project at any time.
                        My refusal to participate will not result in any penalty.
                    </li>
                    <li>
                        I do not waive any legal rights or release Princeton University, its agents, or you from liability for negligence.<br><br>
                    </li>
                </ol>
            </li>
            <li>
                I hereby give my consent to be the subject of your research.
            </li>
        </ol>
    </p>

    <p>
        <span style="font-weight: bold; margin-top: 15px;">
            This study has been approved by the Institutional Review Board for Human Subjects
        </span>
    </p>
    {{ consent.buttons(config) }}
  </div>
</div>
{% endblock %}
