{% extends "timeline-page.html" %}

{% block stylesheets %}
    <link rel="stylesheet" href="/static/css/consent.css"/>
{% endblock %}

{% block main_body %}
    <div class="main-div" style="padding-bottom: 200px;">
        <hr>
        <p>
        <h4 style="color: red; font-weight: bold;">
            {{ pgettext("consent-title", "In this experiment, you may be asked to make a voice or video recording.") }}
            {{ pgettext("consent-title", "For example, you may be asked to record your finger-tapping in response to a sound or video, or to record yourself speaking or singing.") }}
        </h4>
        </p>
        <div class="text-justify">
            <div class="text-left" style="font-size: 18px;">
                {{ consent.confidentiality(config, true) }}
                {{ consent.collaborations(config, true) }}
                {{ consent.security(config, true) }}
                {{ consent.contact(config) }}
                {{ consent.consent(config) }}
            </div>
            {{ consent.fixed_buttons(config) }}
        </div>
    </div>
{% endblock %}
