{% extends "timeline-page.html" %}


{% block stylesheets %}
    <link rel="stylesheet" href="/static/css/consent.css"/>
{% endblock %}

{% block main_body %}
  <div class="main-div" style="padding-bottom: 100px;">

    <h1>{{ pgettext("consent-title", "Research Participant Consent Form") }}</h1>
    <hr>
    <div class="text-justify">
        <div class="text-left" style="font-size:18px">
          <h2>{{ pgettext("consent", "Title of study:") }} {{experiment_title}}</h2>
          <p>
              {{ pgettext("consent-cap", "I have received written information about the study and the procedure of the study.") }}
              {{ pgettext("consent-cap", "I have read and understood all of the information completely.") }}
              {{ pgettext("consent-cap", "If I had questions concerning the upcoming study, they were answered completely and to my satisfaction.") }}
          </p>
          <p>
              {{ pgettext("consent-cap", "I agree to the described collecting and processing of my age, gender, and country of birth.") }}
              {{ pgettext("consent-cap", "The recording and processing of these data are carried out in pseudonymized form by means of a code and do not contain my name.") }}
              {{ pgettext("consent-cap", "There is a coding list on paper that links my name to this number.") }}
              {{ pgettext("consent-cap", "The investigators are the only persons who have access to this coding list and who are thus able to link the collected data to my name.") }}
              {{ pgettext("consent-cap", "After completion of the research project, the coding list will be destroyed.") }}
              {{ pgettext("consent-cap", "From this point on, my data are anonymous, which means that they cannot be linked to my person any longer.") }}
          </p>
          <p>
              {{ pgettext("consent", "I understand that I am free to withdraw my consent to the collection and storage of my personal data without negative consequences.") }}
              {{ pgettext("consent-cap", "I have been informed that I may request deletion of all my data at any time.") }}
              {{ pgettext("consent-cap", "However, if the coding list has already been destroyed, it will no longer be possible to identify and thus to delete my data set.") }}
          </p>
          <p>
              {{ pgettext("consent-cap", "The participant information sheet about the examination is part of this consent form.") }}
              {{ pgettext("consent-cap", "I have received a copy of the participant information sheet and of the consent form.") }}
          </p>
            <h4>{{ pgettext("consent", "Investigator") }}</h4>
          <p>
              Harin Lee,
              Max Planck Institut for Empirical Aesthetics,
              Grüneburgweg 14,
              60322 Frankfurt am Main,
              harin.lee@ae.mpg.de
          </p>
          {{ consent.pi(config) }}
          {{ consent.consent(config) }}
        </div>
        {{ consent.buttons(config) }}
      </div>
  </div>
{% endblock %}
