{% extends "psynet_layout.html" %}

{% block body %}
    <div id="container">
      <div style="max-width: max-content;">
        <div class="col-xs-2">
          <img id="adlogo" src="/static/images/logo.png" alt="Psynet logo" style="width: 150px; float: left;">
        </div>
        <h1 id="header" style="clear: both;">{{ pgettext("abort_possible", "Aborting not possible.") }}</h1>
        <hr>
        <p>
          {{ pgettext("abort_possible", "Aborting the experiment is not possible as the experimenter does not allow it at this time!") }}
            {% set CURRENCY = config.currency %}
            {% set MIN_ACCUMULATED_REWARD_FOR_ABORT = "{:.2f}".format(config.min_accumulated_reward_for_abort) %}
            {% trans name="abort_possible" %}You have to accumulate <strong>at least {{ CURRENCY }}{{MIN_ACCUMULATED_REWARD_FOR_ABORT}}</strong> of reward to be able to be compensated.{% endtrans %}
           {{ pgettext("abort_possible", "See below information on your currently accumulated reward.") }}
        </p>
        <table class="table table-sm table-striped">
            <tbody>
              <tr><th>{{ gettext("Assignment ID") }}</th><td>{{ participant_abort_info.assignment_id }}</td></tr>
              <tr><th>{{ gettext("HIT ID") }}</th><td>{{ participant_abort_info.hit_id }}</td></tr>
              <tr><th>{{ gettext("Accumulated reward") }}</th><td>{{ participant_abort_info.accumulated_reward }}</td></tr>
            </tbody>
        </table>
        <button type="button" id="close-button" class="btn btn-primary btn-lg" style="float: left; margin-top: 15px;" onclick="window.close();">
          {{ gettext("Close window") }}
        </button>
      </div>
    </div>
{% endblock %}

