{% extends "timeline-page.html" %}

{% block stylesheets %}
    <link rel="stylesheet" href="/static/css/consent.css"/>
{% endblock %}

{% block main_body %}
  <div class="main-div" style="padding-bottom: 200px;">
    <h1>{{ pgettext("consent-title", "We need your consent to proceed") }}</h1>
    <hr>
    <p>
        {{ pgettext("consent", "Following open science practices to improve science reusability and replicability, we may want to publish the dataset collected in this experiment as soon as the study is finished.") }}
        {{ pgettext("consent", "This includes audio-visual recordings, such as a recording of your finger-tapping in response to a sound or video, or to record yourself speaking or singing.") }}
        {{ pgettext("consent", "We may also want to publish certain materials as part of the scientific publication, such as a demo video of the task or part of an interview.") }}
    </p>
    <div class="text-justify">
        <div class="text-left" style="font-size:18px">
            <h4>{{ pgettext("consent-confidentiality", "Confidentiality") }}</h4>
            <p class="consent-text">
                {{ pgettext("consent-confidentiality", "Since the audio-visual recordings collected in this experiment may be made public, we cannot guarantee anonymity of your data if you participate in this experiment.") }}
                {{ pgettext("consent-confidentiality", "However, your data will not be directly associated with your name or other identifiers in any way (we will not specify your name when presenting audio/video materials of yours).") }}
            </p>

            <h4>{{ pgettext("consent-security", "Security") }}</h4>
            <p class="consent-text">
                {{ pgettext("consent-security", "Data will be published on open science frameworks/websites such as OSF (https://osf.io/) and will be available to everybody with internet access.") }}

            <h4>{{ pgettext("consent-disclaimer", "Disclaimer") }}</h4>
            <p class="consent-text">
                {{ pgettext("consent-disclaimer", "Since some materials will be published as part of the scientific publication, you will not be able to delete such materials, as they will become public (either immediately or soon after data collection).") }}
                {{ pgettext("consent-disclaimer", "Thus, you may wish to make sure you do not record any materials you do not want to make public.") }}
                {{ pgettext("consent-disclaimer", "By consenting to this experiment you are committing not to use profanity while recording audio or nudity while recording video.") }}
                {{ pgettext("consent-disclaimer", "Although your data will not be associated with personal identifiers, it is possible to link your personal information based on audio-visual recordings") }}.
            </p>

            {{ consent.contact(config) }}

            {{ consent.consent(config) }}
        </div>
        {{ consent.fixed_buttons(config) }}
      </div>
  </div>
{% endblock %}
